@@ -1,9 +1,3 @@
-# Copyright (c) 2010 Jan Urbanski.
-# See LICENSE for details.
-
-"""
-Tests for twisted.enterprise.pgadbapi.
-"""
 import os
 
 try:
@@ -16,7 +10,7 @@ except ImportError:
     except ImportError:
         raise ImportError('no module named psycopg2 or psycopg2ct')
 
-from txpostgres import txpostgres
+from txpostgres import txpostgres, reconnection
 
 from twisted.trial import unittest
 from twisted.internet import defer, error, main, posixbase, reactor, task
@@ -457,6 +451,8 @@ class TxPostgresConnectionTestCase(Psycopg2TestCase):
 
 class _SimpleDBSetupMixin(object):
 
+    connectionClass = txpostgres.Connection
+
     def setUp(self):
         d = self.restoreConnection(None)
         d.addCallback(lambda _: self.conn.cursor())
@@ -473,7 +469,7 @@ class _SimpleDBSetupMixin(object):
         been passed through. Useful as an addBoth handler for tests that
         disconnect from the database.
         """
-        self.conn = txpostgres.Connection()
+        self.conn = self.connectionClass()
         d = self.conn.connect(user=DB_USER, password=DB_PASS,
                               host=DB_HOST, database=DB_NAME)
         return d.addCallback(lambda _: res)
@@ -522,6 +518,15 @@ class TxPostgresManualQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         d.callback(c)
         return d
 
+    def test_cursorKwargs(self):
+        """
+        Cursor methods can be called using keyword arguments.
+        """
+        c = self.conn.cursor()
+        d = c.execute(params=(1, ), query="select %s")
+        return d.addCallback(
+            lambda c: self.assertEquals(c.fetchall(), [(1, )]))
+
     def test_errors(self):
         """
         Errors from the database are reported as failures.
@@ -796,13 +801,64 @@ class TxPostgresQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
             raise unittest.SkipTest("This test fails on versions of Twisted "
                                     "affected by Twisted bug #4539")
 
-        d = self.conn.runQuery("select pg_sleep(5)")
-        reactor.callLater(0, self.conn.close)
+        def _checkConnectionLost(f, testcase):
+            if not isinstance(f, failure.Failure):
+                testcase.fail("connectionLost got non-Failure: %r" % f)
+            if not f.check(error.ConnectionDone):
+                testcase.fail("connectionLost got wrong Failure: %r" % f.value)
 
+        producedCursors = []
+
+        class StrictLosingConnection(txpostgres.Connection):
+            testcase = self
+            seenConnectionLost = False
+
+            def connectionLost(self, reason):
+                txpostgres.Connection.connectionLost(self, reason)
+                self.seenConnectionLost = True
+                _checkConnectionLost(reason, self.testcase)
+
+            def cursor(self, *args, **kwargs):
+                cursor = txpostgres.Connection.cursor(self, *args, **kwargs)
+                producedCursors.append(cursor)
+                return cursor
+
+        class StrictLosingCursor(txpostgres.Cursor):
+            testcase = self
+            seenConnectionLost = False
+
+            def connectionLost(self, reason):
+                txpostgres.Cursor.connectionLost(self, reason)
+                self.seenConnectionLost = True
+                _checkConnectionLost(reason, self.testcase)
+
+        def _runAndClose():
+            # start executing the query and close it in the next reactor cycle
+            d = conn.runQuery("select pg_sleep(5)")
+            reactor.callLater(0, conn.close)
+            return d
+
+        def _waitReactorCycle():
+            # wait one reactor cycle to allow callback to propagate
+            d = defer.Deferred()
+            reactor.callLater(0, d.callback, None)
+            return d
+
+        def _checkSeenConnectionLost():
+            self.assertTrue(conn.seenConnectionLost)
+            for cursor in producedCursors:
+                self.assertTrue(cursor.seenConnectionLost)
+
+        conn = StrictLosingConnection()
+        conn.cursorFactory = StrictLosingCursor
+        d = conn.connect(user=DB_USER, password=DB_PASS,
+                         host=DB_HOST, database=DB_NAME)
+
+        d.addCallback(lambda _: _runAndClose())
         # the query fails with a disconnected error
         d = self.assertFailure(d, error.ConnectionDone)
-        # restore the connection, otherwise all the other tests will fail
-        return d.addBoth(self.restoreConnection)
+        d.addCallback(lambda _: _waitReactorCycle())
+        return d.addCallback(lambda _: _checkSeenConnectionLost())
 
 
 class TxPostgresConnectionPoolTestCase(Psycopg2TestCase):
@@ -1406,3 +1462,68 @@ class TxPostgresNotifyTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         d.addCallback(lambda _: self.assertEquals(
                 cooperator.result, [True]))
         return d.addCallback(lambda _: c.close())
+
+
+class SignallingDetector(reconnection.DeadConnectionDetector):
+
+    died = False
+
+    def startReconnecting(self, f):
+        self.died = True
+        return reconnection.DeadConnectionDetector.startReconnecting(self, f)
+
+
+class ReconnectingConnection(txpostgres.Connection):
+
+    def __init__(self, *args, **kwargs):
+        self.clock = task.Clock()
+
+        detector = SignallingDetector(
+            deathChecker=self.checkDataError,
+            reconnectionIterator=self.sillyIterator,
+            reactor=self.clock)
+
+        kwargs['detector'] = detector
+        txpostgres.Connection.__init__(self, *args, **kwargs)
+
+    def checkDataError(self, f):
+        return f.check(psycopg2.DataError)
+
+    def sillyIterator(self):
+        return [1, 1, 1]
+
+
+class TxPostgresReconnectionTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
+
+    connectionClass = ReconnectingConnection
+
+    def test_reconnection(self):
+        """
+        An error causing the connection to die starts the reconnection process.
+        """
+        recovery_d = defer.Deferred()
+        self.conn.detector.addRecoveryHandler(
+            lambda: recovery_d.callback(None))
+
+        d = self.conn.runOperation('insert into simple values (1)')
+        d.addCallback(lambda _: self.assertFalse(self.conn.detector.died))
+
+        # try inserting text, which will cause a DataError and cause a
+        # reconnection
+        d.addCallback(lambda _: self.conn.runOperation(
+            "insert into simple values ('sometext')"))
+        self.assertFailure(d, psycopg2.DataError)
+        d.addCallback(lambda _: self.assertTrue(self.conn.detector.died))
+
+        # the next query will fail immediately
+        d.addCallback(lambda _: self.conn.runQuery('select 1'))
+        self.assertFailure(d, reconnection.ConnectionDead)
+
+        # after the delay expires, a reconnection attempt will be triggered and
+        # the connection will recover eventually
+        d.addCallback(lambda _: self.conn.clock.advance(1))
+        d.addCallback(lambda _: recovery_d)
+
+        # now the query will no loger fail
+        d.addCallback(lambda _: self.conn.runQuery('select 1'))
+        return d.addCallback(self.assertEquals, [(1, )])
