@@ -14,10 +14,10 @@ has been tested with Twisted 10.2 onward). Alternatively, psycopg2-ctypes_ can
 be used in lieu of Psycopg.
 
 txpostgres tries to present an interface that will be familiar to users of both
-Twisted and Psycopg. It features a :class:`~txpostgres.Cursor` wrapper class
+Twisted and Psycopg. It features a :class:`~txpostgres.txpostgres.Cursor` wrapper class
 that mimics the interface of a Psycopg :psycopg:`cursor <cursor.html#cursor>`
 but returns :d:`Deferred` objects. It also provides a
-:class:`~txpostgres.Connection` class that is meant to be a drop-in replacement
+:class:`~txpostgres.txpostgres.Connection` class that is meant to be a drop-in replacement
 for Twisted's :tm:`adbapi.Connection <enterprise.adbapi.Connection>` with some
 small differences regarding connection establishing.
 
@@ -37,7 +37,7 @@ documentation_
 .. _`asynchronous connections`: http://initd.org/psycopg/docs/advanced.html#async-support
 .. _psycopg2-ctypes: http://pypi.python.org/pypi/psycopg2ct
 .. _source: https://github.com/wulczer/txpostgres
-.. _issue tracker: https://github.com/txpostgres/issues
+.. _issue tracker: https://github.com/wulczer/txpostgres/issues
 .. _documentation: http://txpostgres.readthedocs.org/
 """
 
@@ -55,7 +55,6 @@ from zope.interface import implements
 from twisted.internet import interfaces, main, defer, task
 from twisted.python import failure, log
 
-
 try:
     psycopg2.extensions.POLL_OK
 except AttributeError:
@@ -154,9 +153,9 @@ class _PollingMixin(object):
             ignored.
         :type swallowErrors: bool
 
-        :raise: :exc:`~txpostgres.UnexpectedPollResult` when :meth:`poll`
-            returns a result from outside of the :psycopg:`expected list
-            <extensions.html#poll-constants>`.
+        :raise: :exc:`~txpostgres.txpostgres.UnexpectedPollResult` when
+            :meth:`poll` returns a result from outside of the
+            :psycopg:`expected list <extensions.html#poll-constants>`.
         """
         # This method often gets called from the reactor's doRead/doWrite
         # handlers. Don't callback or errback the polling Deferred here, as
@@ -275,9 +274,9 @@ class Cursor(_PollingMixin):
     asynchronous queries execution, so you need to take care not to execute a
     query while another is still being processed.
 
-    In most cases you should just use the :class:`~txpostgres.Connection`
-    methods that will handle the locking necessary to prevent concurrent query
-    execution.
+    In most cases you should just use the
+    :class:`~txpostgres.txpostgres.Connection` methods that will handle the
+    locking necessary to prevent concurrent query execution.
     """
 
     def __init__(self, cursor, connection):
@@ -360,7 +359,7 @@ class RollbackFailed(Exception):
     state.
 
     :var connection: The connection that failed to roll back its transaction.
-    :vartype connection: :class:`~txpostgres.Connection`
+    :vartype connection: :class:`~txpostgres.txpostgres.Connection`
 
     :var originalFailure: The failure that caused the connection to try to
         roll back the transaction.
@@ -395,16 +394,16 @@ class Connection(_PollingMixin):
     :vartype connectionFactory: any callable
 
     :var cursorFactory: The factory used to produce cursors, defaults to
-        :class:`~txpostgres.Cursor`
+        :class:`~txpostgres.txpostgres.Cursor`
     :vartype cursorFactory: a callable accepting two positional arguments, a
         :psycopg:`psycopg2.cursor <cursor.html#cursor>` and a
-        :class:`~txpostgres.Connection`
+        :class:`~txpostgres.txpostgres.Connection`
     """
 
     connectionFactory = staticmethod(psycopg2.connect)
     cursorFactory = Cursor
 
-    def __init__(self, reactor=None, cooperator=None):
+    def __init__(self, reactor=None, cooperator=None, detector=None):
         if not reactor:
             from twisted.internet import reactor
         if not cooperator:
@@ -413,6 +412,7 @@ class Connection(_PollingMixin):
 
         self.reactor = reactor
         self.cooperator = cooperator
+        self.detector = detector
         self.prefix = "connection"
 
         # this lock will be used to prevent concurrent query execution
@@ -435,9 +435,12 @@ class Connection(_PollingMixin):
 
         :return: A :d:`Deferred` that will fire when the connection is open.
 
-        :raise: :exc:`~txpostgres.AlreadyConnected` when the connection has
-            already been opened.
+        :raise: :exc:`~txpostgres.txpostgres.AlreadyConnected` when the
+            connection has already been opened.
         """
+        if self.detector:
+            self.detector.setReconnectable(self, *args, **kwargs)
+
         if self._connection and not self._connection.closed:
             return defer.fail(AlreadyConnected())
 
@@ -482,9 +485,9 @@ class Connection(_PollingMixin):
         # twice, as the second call will not have any effects.
 
         for cursor in set(self._cursors):
-            cursor.connectionLost(main.CONNECTION_DONE)
+            cursor.connectionLost(failure.Failure(main.CONNECTION_DONE))
 
-        self.connectionLost(main.CONNECTION_DONE)
+        self.connectionLost(failure.Failure(main.CONNECTION_DONE))
 
     def cursor(self):
         """
@@ -497,9 +500,9 @@ class Connection(_PollingMixin):
         Execute an SQL query and return the result.
 
         An asynchronous cursor will be created and its
-        :meth:`~txpostgres.Cursor.execute` method will be invoked with the
-        provided arguments. After the query completes the results will be
-        fetched and the returned :d:`Deferred` will fire with the result.
+        :meth:`~txpostgres.txpostgres.Cursor.execute` method will be invoked
+        with the provided arguments. After the query completes the results will
+        be fetched and the returned :d:`Deferred` will fire with the result.
 
         The connection is always in autocommit mode, so the query will be run
         in a one-off transaction. In case of errors a :tm:`Failure
@@ -511,7 +514,7 @@ class Connection(_PollingMixin):
         :return: A :d:`Deferred` that will fire with the return value of the
             cursor's :meth:`fetchall` method.
         """
-        return self.lock.run(self._runQuery, *args, **kwargs)
+        return self._doit(self._runQuery, *args, **kwargs)
 
     def _runQuery(self, *args, **kwargs):
         c = self.cursor()
@@ -523,17 +526,17 @@ class Connection(_PollingMixin):
         """
         Execute an SQL query and discard the result.
 
-        Identical to :meth:`~txpostgres.Connection.runQuery`, but the result
-        won't be fetched and instead :class:`None` will be returned. It is
-        intended for statements that do not normally return values, like INSERT
-        or DELETE.
+        Identical to :meth:`~txpostgres.txpostgres.Connection.runQuery`, but
+        the result won't be fetched and instead :class:`None` will be
+        returned. It is intended for statements that do not normally return
+        values, like INSERT or DELETE.
 
         It is safe to call this method multiple times without waiting for the
         first query to complete.
 
         :return: A :d:`Deferred` that will fire :class:`None`.
         """
-        return self.lock.run(self._runOperation, *args, **kwargs)
+        return self._doit(self._runOperation, *args, **kwargs)
 
     def _runOperation(self, *args, **kwargs):
         c = self.cursor()
@@ -546,11 +549,11 @@ class Connection(_PollingMixin):
         Run commands in a transaction and return the result.
 
         :obj:`interaction` should be a callable that will be passed a
-        :class:`~txpostgres.Cursor` object. Before calling :obj:`interaction` a
-        new transaction will be started, so the callable can assume to be
-        running all its commands in a transaction. If :obj:`interaction`
-        returns a :d:`Deferred` processing will wait for it to fire before
-        proceeding.
+        :class:`~txpostgres.txpostgres.Cursor` object. Before calling
+        :obj:`interaction` a new transaction will be started, so the callable
+        can assume to be running all its commands in a transaction. If
+        :obj:`interaction` returns a :d:`Deferred` processing will wait for it
+        to fire before proceeding.
 
         After :obj:`interaction` finishes work the transaction will be
         automatically committed. If it raises an exception or returns a
@@ -561,10 +564,11 @@ class Connection(_PollingMixin):
         the failure obtained trying to commit will be returned.
 
         If rolling back the transaction fails the failure obtained from the
-        rollback attempt will be logged and a :exc:`~txpostgres.RollbackFailed`
-        failure will be returned. The returned failure will contain references
-        to the original failure that caused the transaction to be rolled back
-        and to the :class:`~txpostgres.Connection` in which that happend, so
+        rollback attempt will be logged and a
+        :exc:`~txpostgres.txpostgres.RollbackFailed` failure will be
+        returned. The returned failure will contain references to the original
+        failure that caused the transaction to be rolled back and to the
+        :class:`~txpostgres.txpostgres.Connection` in which that happend, so
         the user can take a decision whether she still wants to be using it or
         just close it, because an open transaction might have been left open in
         the database.
@@ -573,14 +577,13 @@ class Connection(_PollingMixin):
         first query to complete.
 
         :param interaction: A callable whose first argument is a
-            :class:`~txpostgres.Cursor`.
+            :class:`~txpostgres.txpostgres.Cursor`.
         :type interaction: any callable
 
         :return: A :d:`Deferred` that will fire with the return value of
             :obj:`interaction`.
         """
-        return self.lock.run(
-            self._runInteraction, interaction, *args, **kwargs)
+        return self._doit(self._runInteraction, interaction, *args, **kwargs)
 
     def _runInteraction(self, interaction, *args, **kwargs):
         c = self.cursor()
@@ -611,6 +614,13 @@ class Connection(_PollingMixin):
 
         return d
 
+    def _doit(self, method, *args, **kwargs):
+        if self.detector:
+            args = (method, ) + args
+            method = self.detector.callChecking
+
+        return self.lock.run(method, *args, **kwargs)
+
     def cancel(self, d):
         """
         Cancel the current operation. The cancellation does not happen
@@ -621,7 +631,7 @@ class Connection(_PollingMixin):
         really be sure which one are you cancelling.
 
         :param d: a :d:`Deferred` returned by one of
-            :class:`~txpostgres.Connection` methods.
+            :class:`~txpostgres.txpostgres.Connection` methods.
         """
         try:
             d.cancel()
@@ -630,9 +640,9 @@ class Connection(_PollingMixin):
 
     def cursorRunning(self, cursor):
         """
-        Called automatically when a :class:`~txpostgres.Cursor` created by this
-        :class:`~txpostgres.Connection` starts polling after executing a query. User
-        code should never have to call this method.
+        Called automatically when a :class:`~txpostgres.txpostgres.Cursor` created
+        by this :class:`~txpostgres.txpostgres.Connection` starts polling after
+        executing a query. User code should never have to call this method.
         """
         # The cursor will now proceed to poll the psycopg2 connection, so stop
         # polling it ourselves until it's done. Failure to do so would result
@@ -644,9 +654,10 @@ class Connection(_PollingMixin):
 
     def cursorFinished(self, cursor):
         """
-        Called automatically when a :class:`~txpostgres.Cursor` created by this
-        :class:`~txpostgres.Connection` is done with polling after executing a
-        query. User code should never have to call this method.
+        Called automatically when a :class:`~txpostgres.txpostgres.Cursor` created
+        by this :class:`~txpostgres.txpostgres.Connection` is done with polling
+        after executing a query. User code should never have to call this
+        method.
         """
         self._cursors.remove(cursor)
         # The cursor is done polling, resume watching the connection for NOTIFY
@@ -672,6 +683,12 @@ class Connection(_PollingMixin):
         if not self._connection.closed:
             self.reactor.addReader(self)
 
+    def connectionLost(self, reason):
+        _PollingMixin.connectionLost(self, reason)
+
+        if self.detector:
+            self.detector.checkForDeadConnection(reason)
+
     def checkForNotifies(self):
         """
         Check if :pg:`NOTIFY <notify>` events have been received and if so,
@@ -745,15 +762,15 @@ class Connection(_PollingMixin):
 
 class ConnectionPool(object):
     """
-    A poor man's pool of :class:`~txpostgres.Connection` instances.
+    A poor man's pool of :class:`~txpostgres.txpostgres.Connection` instances.
 
     :var min: The amount of connections that will be open when
-       :meth:`.start` is called. The pool never opens or closes connections on
-       its own after starting. Defaults to 3.
+        :meth:`~ConnectionPool.start` is called. The pool never opens or closes
+        connections on its own after starting. Defaults to 3.
     :vartype min: int
 
     :var connectionFactory: The factory used to produce connections, defaults
-        to :class:`~txpostgres.Connection`.
+        to :class:`~txpostgres.txpostgres.Connection`.
     :vartype connectionFactory: any callable
 
     :var reactor: The reactor passed to :attr:`.connectionFactory`.
@@ -829,7 +846,8 @@ class ConnectionPool(object):
         pending.
 
         :param connection: The connection to be removed.
-        :type connection: an object produced by the pool's :attr:`connectionFactory`
+        :type connection: an object produced by the pool's
+            :attr:`connectionFactory`
         """
         if not self.connections:
             raise ValueError("Connection still in use")
@@ -860,11 +878,11 @@ class ConnectionPool(object):
         Execute an SQL query using a pooled connection and return the result.
 
         One of the pooled connections will be chosen, its
-        :meth:`~txpostgres.Connection.runQuery` method will be called and the
-        resulting :d:`Deferred` will be returned.
+        :meth:`~txpostgres.txpostgres.Connection.runQuery` method will be
+        called and the resulting :d:`Deferred` will be returned.
 
         :return: A :d:`Deferred` obtained by a pooled connection's
-            :meth:`~txpostgres.Connection.runQuery`
+            :meth:`~txpostgres.txpostgres.Connection.runQuery`
         """
         return self._semaphore.run(self._runQuery, *args, **kwargs)
 
@@ -878,11 +896,11 @@ class ConnectionPool(object):
         Execute an SQL query using a pooled connection and discard the result.
 
         One of the pooled connections will be chosen, its
-        :meth:`~txpostgres.Connection.runOperation` method will be called and
-        the resulting :d:`Deferred` will be returned.
+        :meth:`~txpostgres.txpostgres.Connection.runOperation` method will be
+        called and the resulting :d:`Deferred` will be returned.
 
         :return: A :d:`Deferred` obtained by a pooled connection's
-            :meth:`~txpostgres.Connection.runOperation`
+            :meth:`~txpostgres.txpostgres.Connection.runOperation`
         """
         return self._semaphore.run(self._runOperation, *args, **kwargs)
 
@@ -897,15 +915,17 @@ class ConnectionPool(object):
         result.
 
         One of the pooled connections will be chosen, its
-        :meth:`~txpostgres.Connection.runInteraction` method will be called and
-        the resulting :d:`Deferred` will be returned.
+        :meth:`~txpostgres.txpostgres.Connection.runInteraction` method will be
+        called and the resulting :d:`Deferred` will be returned.
 
         :param interaction: A callable that will be passed to
-            :meth:`Connection.runInteraction <txpostgres.Connection.runInteraction>`
+            :meth:`Connection.runInteraction
+            <txpostgres.Connection.runInteraction>`
         :type interaction: any callable
 
         :return: A :d:`Deferred` obtained by a pooled connection's
-            :meth:`Connection.runInteraction <txpostgres.Connection.runInteraction>`
+            :meth:`Connection.runInteraction
+            <txpostgres.Connection.runInteraction>`
         """
         return self._semaphore.run(
             self._runInteraction, interaction, *args, **kwargs)
